/*----------------------------------------------------------------------------
/ USBN9604											2005 (C)LogicPack
/	05.07.22					000		M.Yamada
/		first release(use M-802)
/---------------------------------------------------------------------------*/
#include <stdio.h>
#include "iodefine.h"
#include "usbn9604io.h"
#include "descriptor.h"
#include "usbn9604.h"

/*-----------------------------------------------
/ *DEBUG
/----------------------------------------------*/
//#define DEBUG_LOG_INTR

#ifdef DEBUG_LOG_INTR
  #define SETLOG_EVENT(d)	(usb_log_buf[usb_log_count].event = d)
  #define SETLOG_SUB(d)		(usb_log_buf[usb_log_count].sub = d)
  #define SETLOG_RXSTAT(d)	(usb_log_buf[usb_log_count].rxstat = d)
  #define SETLOG_UB0(d)		(usb_log_buf[usb_log_count].ub0 = d)
  #define SETLOG_UB1(d)		(usb_log_buf[usb_log_count].ub1 = d)
  #define SETLOG_UB2(d)		(usb_log_buf[usb_log_count].ub2 = d)
  #define SETLOG_UB3(d)		(usb_log_buf[usb_log_count].ub3 = d)
  #define SETLOG_UB4(d)		(usb_log_buf[usb_log_count].ub4 = d)
  #define SETLOG_UB5(d)		(usb_log_buf[usb_log_count].ub5 = d)
#else
  #define SETLOG_EVENT(d)
  #define SETLOG_SUB(d)
  #define SETLOG_RXSTAT(d)
  #define SETLOG_UB0(d)
  #define SETLOG_UB1(d)
  #define SETLOG_UB2(d)
  #define SETLOG_UB3(d)
  #define SETLOG_UB4(d)
  #define SETLOG_UB5(d)
#endif

/*-----------------------------------------------
/ *DEFINE
/----------------------------------------------*/
#define USB_CLKDIV		0x04		/* CLKOUT = 48MHz/4 = 12MHz */

typedef struct{
	volatile byte *rp;				/* read pointer */
	volatile byte *wp;				/* write pointer */
	int size;						/* buffer size */
	byte *buffer;					/* buffer pointer */
	bool full;						/* buffer full flag */
	bool addition;					/* buffer ring flag */
	ENDPOINT_INFO endpoint;
}BUF_CTL;

/*-----------------------------------------------
/ *STATIC
/----------------------------------------------*/
#if defined(DEBUG_LOG_INTR)
#pragma section EXT
USB_LOG usb_log_buf[USB_LOG_MAX];
int usb_log_count;
int usb_intr_count;
#pragma section
#endif

static const byte RXC[4] = { RXC0, RXC1, RXC2, RXC3 };
static const byte EPC[8] = { EPC0, EPC1, EPC2, EPC3, EPC4, EPC5, EPC6, EPC0 };
static const byte TXC[4] = { TXC0, TXC1, TXC2, TXC3 };

static byte SETADDR;					/* AhXZbg 			*/
static byte configno;					/* RtBO[VNO	*/
static byte usbbuff[64];				/* ǂݍ݃obt@			*/
static byte STALLD;						/* ECP̏				*/
static byte DATA0_1;					/* USB_TXTGL̃tO		*/

static struct{
	bool send;							/* 1 = fBXNv^M	*/
	int size;							/* fBXNv^MTCY	*/
	char *ptr;							/* fBXNv^|C^	*/
}desc_ctrl;

static BUF_CTL tx[4];
static BUF_CTL rx[4];

/* efXNv^ւ̃|C^ */
static const byte *dev_desc;
static const byte *cfg_desc;
static const byte *lang_data;
static const byte *mfg_str;
static const byte *nbr_str;
static const byte *int_str;

static const SETTING_INFO *setting;		/*  */

/*---------------------------------*/
static void setup_reg(void);

static void exec_rx(int no);
static void exec_rx0(void);
static void exec_tx0(void);
static void exec_tx(int no);
static void nako0(void);
static void nako1(void);
static void naki0(void);
static void naki1(void);
static void exec_request(void);

static void clrfeature(void);
static void setfeature(void);
static void getdescriptor(void);
static void exec_set_addr(void);
static void exec_set_if(void);
static void send_desc_sub(void *ptr,int size);
static void send_desc(void);
static void getstatus(void);
static void setconfiguration(void);

static void set_stall(int epcnum);
static void clear_stall(int epcnum);

#define rx_flush(n)		write_reg(RXC[n], USB_FLUSH)
#define rx_enable(n)	write_reg(RXC[n], USB_RX_EN)

static void tx_flush(int no);
static void tx_togle(int no);
static void tx_enable(int no, bool last);

/* HID */
static void setprotocol(void);
static void getprotocol(void);
static void setidle(void);
static void getidle(void);
static void setreport(void);
static void getreport(void);

/* I/O */
static byte read_reg(byte adr);
static void write_reg(byte adr, byte data);
static int read_reg_burst(byte adr, byte adr2, byte *buff, int cnt);
static int write_reg_burst(byte adr, const byte *buff, int cnt);

/*-------------------------------------------------------------------
/ ݒ
/	ENT:
/	  DESCRIPTOR_INFO *desc
/	  SETTING_INFO *set
/------------------------------------------------------------------*/
void UsbSetAction(const DESCRIPTOR_INFO *desc, const SETTING_INFO *set)
{
	/* fXNv^ւ̃|C^ێ */
	dev_desc = desc->device;
	cfg_desc = desc->config;
	lang_data = desc->langage;
	mfg_str = desc->mfg_str;
	nbr_str = desc->nbr_str;
	int_str = desc->int_str;
	/* ݒRs[ */
	setting = set;
}
/*-------------------------------------------------------------------
/ obt@ݒ
/	ENT:
/	  USBBUF_INFO *buf
/	  ENDPOINT_INFO *endp
/	DOC:
/	  additionTRUEɂꍇAbuf̓Oobt@
/	  ƂėpB
/	  additionFALSEɂꍇf[^̓obt@̐擪ɂȂ炸
/	  ܂
/------------------------------------------------------------------*/
void UsbSetWriteBuf(const USBBUF_INFO *buf, const ENDPOINT_INFO *endp)
{
	tx[buf->no].size = buf->size;
	tx[buf->no].buffer = buf->tmp;
	tx[buf->no].rp = buf->tmp;
	tx[buf->no].wp = buf->tmp;
	tx[buf->no].addition = buf->addition;
	tx[buf->no].endpoint = *endp;
}

/*-------------------------------------------------------------------
/ Mobt@֏
/	ENT:
/	  int no									Gh|C
/	  byte *p									obt@|C^
/	  int size									݃TCY
/	RET:
/	  int										񂾃TCY
/------------------------------------------------------------------*/
int UsbPut(int no, const byte *p,int size)
{
	byte *wp;
	int max = 0;
	if( tx[no].full ){ return 0; }
	if( !tx[no].addition ){
		if( tx[no].wp != tx[no].rp ){
			return 0;
		}	/* IĂȂ */
	}
	while( size > 0 ){
		*tx[no].wp = *p;
		wp = tx[no].wp + 1;
		if( wp == &tx[no].buffer[tx[no].size] ){
			wp = &tx[no].buffer[0];
		}
		tx[no].wp = wp;
		max++;
		size--;
		p++;
		if( tx[no].wp == tx[no].rp ){
			tx[no].full = TRUE;
			break;
		}
	}
	return max;
}

/*-------------------------------------------------------------------
/ obt@ݒ
/	ENT:
/	  USBBUF_INFO *buf					
/	  ENDPOINT_INFO *endp				
/------------------------------------------------------------------*/
void UsbSetReadBuf(const USBBUF_INFO *buf, const ENDPOINT_INFO *endp)
{
	rx[buf->no].size = buf->size;
	rx[buf->no].buffer = buf->tmp;
	rx[buf->no].rp = buf->tmp;
	rx[buf->no].wp = buf->tmp;
	rx[buf->no].addition = buf->addition;
	rx[buf->no].endpoint = *endp;
}

/*-------------------------------------------------------------------
/ f[^擾
/	ENT:
/	  int no							Gh|Cg
/	  byte *p							obt@ւ̃|C^
/	  int size;							obt@TCY
/	RET:
/	  int								ۓǂݍ񂾃TCY
/------------------------------------------------------------------*/
int UsbGet(int no, byte *p,int size)
{
	int len = 0;
	byte *rp;
	if( !rx[no].full ){
		if( rx[no].rp == rx[no].wp ){ return 0; }
	}
	while( size > 0 ){
		*p = *rx[no].rp;
		rp = rx[no].rp + 1;
		if( rp == &rx[no].buffer[rx[no].size] ){
			rp = &rx[no].buffer[0];
		}
		rx[no].rp = rp;
		len++;
		size--;
		p++;
		if( rx[no].rp == rx[no].wp ){ break; }
	}

	if( rx[no].full ){
		rx[no].full = FALSE;
	}
	return len;
}

/*-------------------------------------------------------------------
/ f[^擾
/	ENT:
/	  int no
/	RET:
/	  int								f[^TCY
/------------------------------------------------------------------*/
int UsbGetBufLen(int no)
{
	int len;
	if(rx[no].wp >= rx[no].rp){
		len = rx[no].wp - rx[no].rp;
	}else{
		len = rx[no].size - (rx[no].rp - rx[no].wp);
	}
	return len;
}

/*-------------------------------------------------------------------
/	
/------------------------------------------------------------------*/
void UsbInit(void)
{
	UsbReset();
	write_reg(MCNTRL, INTLP | VGE);		/* ݂active low push pull */
	write_reg(CCONF, USB_CLKDIV-1);		/* 48MHz/4 = 12MHz */

	setup_reg();
}

/*-------------------------------------------------------------------
/	Zbg
/------------------------------------------------------------------*/
void UsbReset(void)
{
	write_reg(MCNTRL,SRST);				/* \tgEFAZbg */
	while(read_reg(MCNTRL) & SRST);		/* Zbg҂ */
}

/*-------------------------------------------------------------------
/	USB荞
/------------------------------------------------------------------*/
#if defined(RAM_EMULATE) || defined(INTR_INDIRECT)
void usb_intr(void)
#else
#pragma section IntPRG
__interrupt(vect=17) void usb_intr(void)
#endif
{
	byte usbevent;
	byte sub;
	#define nakevent	sub
	#define rxevent		sub
	#define txevent		sub
	#define altevent	sub

	if( INTC.ISR.BIT.IRQ5F ){
		INTC.ISR.BIT.IRQ5F = 0;
		usbevent = read_reg(MAEV);

		SETLOG_EVENT(usbevent);										//DEBUG
		SETLOG_SUB(0x00);											//DEBUG
		SETLOG_RXSTAT(0x00);										//DEBUG
		SETLOG_UB0(0x00);											//DEBUG
		SETLOG_UB1(0x00);											//DEBUG
		SETLOG_UB2(0x00);											//DEBUG
		SETLOG_UB3(0x00);											//DEBUG
		SETLOG_UB4(0x00);											//DEBUG
		SETLOG_UB5(0x00);											//DEBUG

		usbevent &= (NAK | RX_EV | TX_EV | ALT);

		while( usbevent ){
			if( usbevent & NAK ){
				nakevent = read_reg(NAKEV);
				if( nakevent & NAK_O0 ){ nako0(); }
				if( nakevent & NAK_O1 ){ nako1(); }
				if( nakevent & NAK_I0 ){ naki0(); }
				if( nakevent & NAK_I1 ){ naki1(); }
				SETLOG_EVENT(NAK);									//DEBUG
				SETLOG_SUB(nakevent);								//DEBUG
				usbevent &= ~NAK;
			}
			else if( usbevent & RX_EV ){
				rxevent = read_reg(RXEV);
				if( rxevent & RXFIFO0 ){ exec_rx0(); }
				if( rxevent & RXFIFO1 ){ exec_rx(1); }
				if( rxevent & RXFIFO2 ){ exec_rx(2); }
				if( rxevent & RXFIFO3 ){ exec_rx(3); }
				SETLOG_EVENT(RX_EV);								//DEBUG
				SETLOG_SUB(rxevent);								//DEBUG
				usbevent &= ~RX_EV;
			}
			else if( usbevent & TX_EV ){
				txevent = read_reg(TXEV);
				if( txevent & TXFIFO0 ){ exec_tx0(); }
				if( txevent & TXFIFO1 ){ exec_tx(1); }
				if( txevent & TXFIFO2 ){ exec_tx(2); }
				if( txevent & TXFIFO3 ){ exec_tx(3); }
				SETLOG_EVENT(TX_EV);								//DEBUG
				SETLOG_SUB(txevent);								//DEBUG
				usbevent &= ~TX_EV;
			}
			else if( usbevent & ALT ){
				altevent = read_reg(ALTEV);
				if( altevent & ALT_RESET ){				/* Zbg */
					setup_reg();
//					usb_wakeup();
					altevent = ALT_RESET;							//DEBUG
				}
				else if( altevent & ALT_SD3 ){				/* TXyh */
					write_reg(ALTMSK,ALT_RESUME+ALT_RESET);	/* ALTMSK		*/
					write_reg(NFSR,SUS_ST);						/* ɰނ	*/
					altevent = ALT_SD3;								//DEBUG
				}
				else if( altevent & ALT_RESUME ){			/* W[ */
					write_reg(ALTMSK,ALT_SD3+ALT_RESET);	/* ALTMSK			*/
					write_reg(NFSR,OPR_ST);				/* ɰނ𓮍\ɂ	*/
					altevent = ALT_RESUME;							//DEBUG
				}
				SETLOG_EVENT(ALT);									//DEBUG
				SETLOG_SUB(altevent);								//DEBUG
				usbevent &= ~ALT;
			}
#ifdef DEBUG_LOG_INTR												//DEBUG
			if( usb_log_buf[usb_log_count].sub != 0 ){
				if( ++usb_log_count == USB_LOG_MAX ){
					usb_log_count = 0;
				}
			}
#endif
		}
#ifdef DEBUG_LOG_INTR												//DEBUG
		usb_intr_count++;
#endif
	}
}

/* eWX^ */
static void setup_reg(void)
{
	STALLD = 0;
	desc_ctrl.send = 0;
	DATA0_1 = 0;
	SETADDR = 0;

	write_reg(FAR, AD_EN+0);							/* AhX */
	write_reg(EPC0, USB_EP_EN);							/* EP0Enable */
	write_reg(NAKMSK, NAK_O0);
	write_reg(TXMSK, TXFIFO0+TXFIFO1+TXFIFO2+TXFIFO3);
	write_reg(RXMSK, RXFIFO0+RXFIFO1+RXFIFO2+RXFIFO3);
	write_reg(ALTMSK, ALT_SD3+ALT_RESET);
	write_reg(MAMSK, INTR_E+RX_EV+NAK+TX_EV+ALT);
	tx_flush(0);

	tx_flush(1);	write_reg(TXC1,0);
	tx_flush(2);	write_reg(TXC2,0);
	tx_flush(3);	write_reg(TXC3,0);
	rx_flush(1);	write_reg(RXC1,0);
	rx_flush(2);	write_reg(RXC2,0);
	rx_flush(3);	write_reg(RXC3,0);

	write_reg(RXC0, USB_RX_EN);				/* RX0Enable */
	write_reg(NFSR, OPR_ST);				/* ɂ */
	write_reg(MCNTRL, INTLP | NAT | VGE);	/* USBm[h𓮍ɂ */
}

/*------------------------------------------------------------*/
/* RX0 M */
static void exec_rx0(void)
{
	byte rxstat;
	rxstat = read_reg(RXS0);
	if( rxstat & USB_SETUP_RX ){
		read_reg_burst(RXD0,RXS0,usbbuff,8);
		rx_flush(0);
		tx_flush(0);
		clear_stall(0);
		exec_request();
		DATA0_1 |= 1;	/* SETUPȂ̂ް̗LɊ֌WDATA1ƂđM */
		tx_togle(0);
		SETLOG_UB0(usbbuff[0]);										//DEBUG
	}else{
		if( desc_ctrl.send ){
			desc_ctrl.send = 0;
		}
		tx_flush(0);
		write_reg(RXC0,USB_RX_EN);
	}
	SETLOG_RXSTAT(rxstat);											//DEBUG
}

static void exec_request(void)
{
	byte request;
	request = usbbuff[0] & REQUEST_TYPE;
	/* WfoCXNGXg */
	if( request == RTYPE_STD ){
		switch( usbbuff[1] ){
			case CLEAR_FEATURE:		clrfeature();				break;
			case GET_CONFIGURATION:	write_reg(TXD0,configno);	break;
			case GET_DESCRIPTOR:	getdescriptor();			break;
			case GET_STATUS:		getstatus();				break;
			case GET_INTERFACE:		write_reg(TXD0,0);			break;
			case SET_ADDRESS:		exec_set_addr();			break;
			case SET_CONFIGURATION:	setconfiguration();			break;
			case SET_FEATURE:		setfeature();				break;
			case SET_INTERFACE:		exec_set_if();				break;
			default:				set_stall(0)	;			break;		/* ` */
		}
		SETLOG_UB1(usbbuff[1]);									//DEBUG
	}
	/* NXNGXg */
	else if( request == RTYPE_CLASS ){
		if( !setting->class_request ){
			set_stall(0);
		}else{
			switch( usbbuff[1] ){
				case GET_REPORT:	getreport();		break;
				case GET_IDLE:		getidle();			break;
				case GET_PROTOCOL:	getprotocol();		break;
				case SET_REPORT:	setreport();		break;
				case SET_IDLE:		setidle();			break;
				case SET_PROTOCOL:	setprotocol();		break;
				default:			set_stall(0);		break;
			}
		}
	}
	/* x_NGXg */
	else if( request == RTYPE_VENDOR ){
		set_stall(0);
	}
	/* ` */
	else{
		set_stall(0);
	}
}

/*------------------------------------------------------------*/
/* RXx M */
static void exec_rx(int no)
{
	const byte RXS[4] = { RXS0, RXS1, RXS2, RXS3 };
	const byte RXD[4] = { RXD0, RXD1, RXD2, RXD3 };

	static byte temp[64];
	byte rxstat;
	rxstat = read_reg(RXS[no]);
	if( !(rxstat & (USB_SETUP_RX|USB_RX_ERR)) && !rx[no].full ){
		int i, cnt;
		byte *p;
		cnt = read_reg_burst(RXD[no],RXS[no],temp, 64);	/* FIFOf[^擾 */
		p = temp;
		for(i = 0; i < cnt; i++){
			*rx[no].wp = *p;
			if( ++rx[no].wp == &rx[no].buffer[rx[no].size] ){
				rx[no].wp = &rx[no].buffer[0];
			}
			if( rx[no].wp == rx[no].rp ){
				rx[no].full = TRUE;
				break;
			}
			p++;
		}
	}
	rx_flush(no);
	rx_enable(no);
}

/*------------------------------------------------------------*/
/* TX0 M */
static void exec_tx0(void)
{
	byte	txstat;
	txstat = read_reg(TXS0);
	if( (txstat & ACK_STAT) && (txstat & TX_DONE) ){
		/* ok */
		tx_flush(0);
		if( desc_ctrl.send ){
			send_desc();
			tx_togle(0);					/* TX0M */
		}
		else{
			write_reg(RXC0,USB_RX_EN);	/* RX0M */
		}
	}else{
		/* error ? */
	}
}
/*------------------------------------------------------------*/
/* TXx M */
static void exec_tx(int no)
{
	const byte TXS[4] = { TXS0, TXS1, TXS2, TXS3 };
	const byte TXD[4] = { TXD0, TXD1, TXD2, TXD3 };
	static byte temp[64];
	byte txstat;
	txstat = read_reg(TXS[no]);
	if( txstat & TX_DONE ){
		if( txstat & ACK_STAT ){
			int i;
			i = 0;
			tx_flush(no);
			USB9604R = TXD[no];
			while( tx[no].full || (tx[no].rp != tx[no].wp) ){
				USB9604D = *tx[no].rp;
				if(++tx[no].rp == &tx[no].buffer[tx[no].size]){
					tx[no].rp = &tx[no].buffer[0];
				}
				if( ++i >= 64 ){ break; }
			}
			tx[no].full = FALSE;
			if( !tx[no].addition && (tx[no].rp == tx[no].wp) ){
				tx[no].rp = &tx[no].buffer[0];
				tx[no].wp = &tx[no].buffer[0];
			}
			tx_togle(no);
		}else{
			byte d;
			d = USB_TX_EN;
#if 0
			if( DATA0_1 & (1<<no) ){ d |= USB_TX_TOGL; }
			else{					d &= ~USB_TX_TOGL; }
#else
			if( DATA0_1 & (1<<no) ){ d &= ~USB_TX_TOGL; }	/* ȑȌ */
			else{					 d |= USB_TX_TOGL;  }
#endif
			d |= USB_TX_LAST;
			d |= USB_RFF;
			write_reg(TXC[no], d);
//			DATA0_1 ^= (1<<no);
		}
	}
}

/*------------------------------------------------------------*/
/* NAK */
static void nako0(void)
{
}
static void nako1(void)
{
}
static void naki0(void)
{
}

static void naki1(void)
{
}

/*-------------------------------------------------------------------
/ WfoCXNGXg
/------------------------------------------------------------------*/
/* I@ */
static void clrfeature(void)
{
	if( (usbbuff[0] & 0x03) == 0x02 ){	/* Gh|Cgf[^ */
		if((usbbuff[3] & 0x07) != 0 ){
			clear_stall(usbbuff[3] & 0x07);
		}
		STALLD &= -1 ^ (1<<(usbbuff[3] & 7));
	}
}
static void setfeature(void)
{
	if( (usbbuff[0] & 0x03) == 0x02 ){	/* Gh|Cgf[^ */
		if((usbbuff[3] & 0x07) != 0 ){
			set_stall(usbbuff[3] & 0x07);
		}
		STALLD |= (1 << (usbbuff[3] & 0x07));
	}
}

/* fBXNv^Ԃ */
static void getdescriptor(void)
{
	DATA0_1 &= 0xfe;
	switch( usbbuff[3] ){
		case USB_DEVICE:		send_desc_sub(dev_desc, dev_desc[0]);	break;
		case USB_CONFIGURATION:	send_desc_sub(cfg_desc,cfg_desc[2]);	break;
		case USB_HID:			break;
		case USB_HIDREPORT:
			if( setting->report_desc != NULL ){
				send_desc_sub(&setting->report_desc[1], setting->report_desc[0]);
			}
			break;
		case USB_XSTRING:
		{
			switch( usbbuff[2] ){
				case 0: send_desc_sub(lang_data,lang_data[0]);	break;
				case 1: send_desc_sub(mfg_str,mfg_str[0]);		break;
				case 2: send_desc_sub(nbr_str,nbr_str[0]);		break;
				case 3: send_desc_sub(int_str,int_str[0]);		break;
			}
			break;
		}
		default :
		{
		}
	}
	SETLOG_UB2(usbbuff[2]);											//DEBUG
	SETLOG_UB3(usbbuff[3]);											//DEBUG
	SETLOG_UB4(usbbuff[4]);											//DEBUG
	SETLOG_UB5(usbbuff[5]);											//DEBUG
}

static void send_desc_sub(void *ptr,int size)
{
	desc_ctrl.size = (usbbuff[7] << 8) + usbbuff[6];
	if( desc_ctrl.size > size ){		/* Mvobt@ȏ̓f[^𑗂Ȃ */
		desc_ctrl.size = size;
	}
	desc_ctrl.ptr = ptr;
	desc_ctrl.send = 1;								/* fBXNv^MtO𗧂Ă */
	send_desc();
}

static void send_desc(void)
{
	int size;
	size = 8;
	if( desc_ctrl.size == 0 ){ return; }
	if( desc_ctrl.size <= 8 ){ size = desc_ctrl.size; }
	size = write_reg_burst(TXD0, (byte*)desc_ctrl.ptr, size);
	desc_ctrl.size -= size;
	desc_ctrl.ptr += size;
	if( desc_ctrl.size == 0 ){ desc_ctrl.send = 0; }
}

/* foCXAhXv */
static void exec_set_addr(void)
{
	write_reg(EPC0, USB_DEF);
	SETADDR = usbbuff[2] | AD_EN;
	write_reg(FAR,SETADDR);
}

static void exec_set_if(void)
{
	if( !setting->set_interface ){
		if( usbbuff[2] != 0 ){ set_stall(0); }
	}else{
		#define wIndex_low	usbbuff[4]
		#define wValue_low	usbbuff[2]
		if( wIndex_low == 0x01 ){
			if( wValue_low == 0 ){
			}else{
			}
		}
	}
	SETLOG_UB2(usbbuff[2]);											//DEBUG
	SETLOG_UB3(usbbuff[3]);											//DEBUG
	SETLOG_UB4(usbbuff[4]);											//DEBUG
	SETLOG_UB5(usbbuff[5]);											//DEBUG
}

/*------------------------------------------------------------*/
/* Xe[^X */
static void getstatus(void)
{
	int		data,ep;
	data = usbbuff[0] & 3;
	if( (data ==  0) || (data == 1) ){		/* DEVICE,INTERFACE */
		write_reg(TXD0,0);
		write_reg(TXD0,0);
	}else if( data == 2 ){					/* Gh|Cg */
		ep = usbbuff[3] & 7;
		/* epSTALLԂ𑗐M */
		if( STALLD & (1<<ep) )	write_reg(TXD0,1);
		else					write_reg(TXD0,0);
	}else{
		write_reg(TXD0,0);
	}
}

/*------------------------------------------------------------*/
static void setconfiguration(void)
{
	configno = usbbuff[2];
	if( configno == 0 ){
		write_reg(EPC1,0);				/* EPC1gps */
		write_reg(EPC2,0);				/* EPC2gps */
		write_reg(EPC3,0);				/* EPC3gps */
		write_reg(EPC4,0);				/* EPC4gps */
		write_reg(EPC5,0);				/* EPC5gps */
		write_reg(EPC6,0);				/* EPC6gps */
	}else{
		int i;
		STALLD = 0;
		for( i = 1; i <= 3; i++ ){
			if( tx[i].buffer != NULL ){
				tx_flush(i);
				if( !tx[i].endpoint.iso ){
					write_reg(EPC[tx[i].endpoint.epc],
								USB_EP_EN + tx[i].endpoint.addr);
					write_reg(TXC[i], USB_TX_EN|USB_TX_LAST);
				}else{							/* AC\NiX */
					write_reg(EPC[tx[i].endpoint.epc],
								(USB_EP_EN | USB_ISO) + tx[i].endpoint.addr);
					write_reg(TXC[i], USB_TX_EN|USB_TX_LAST|USB_IGNIOS);
				}
			}
			if( rx[i].buffer != NULL ){
				rx_flush(i);
				if( rx[i].endpoint.iso ){
					write_reg(EPC[rx[i].endpoint.epc],
								(USB_EP_EN | USB_ISO) + rx[i].endpoint.addr);
				}else{							/* AC\NiX */
					write_reg(EPC[rx[i].endpoint.epc],
								USB_EP_EN + rx[i].endpoint.addr);
				}
				write_reg(RXC[i], USB_RX_EN);
			}
		}
	}
}

/*-----------------------------------------------
/ HIDNX^NGXg
/----------------------------------------------*/
static byte DESC_TYPE;
static byte USB_IDLE;
static byte USB_PROT;

static void getreport(void)
{
	DESC_TYPE = usbbuff[3];
}
static void setreport(void)
{
	DESC_TYPE = usbbuff[3];
}

static void getidle(void)
{
	write_reg(TXD0,USB_IDLE);
}

static void setidle(void)
{
	USB_IDLE = usbbuff[3];
}

static void getprotocol(void)
{
	write_reg(TXD0,USB_PROT);
}

static void setprotocol(void)
{
	USB_PROT = usbbuff[2];
}

/*----- ėpfunction ------------------------------------*/
/* STALL̃Zbg/NA */
static void set_stall(int epcnum)
{
	write_reg(EPC[epcnum], read_reg(EPC[epcnum]) | STALL);
}

static void clear_stall(int epcnum)
{
	write_reg(EPC[epcnum], read_reg(EPC[epcnum]) & ~STALL);
}

/* Mobt@NA */
static void tx_flush(int no)
{
	int d;
	d = read_reg(TXC[no]);
	d |= USB_FLUSH;
	write_reg(TXC[no],d);
}

/* MItOZbg */
static void tx_togle(int no)
{
	byte d;
	d = USB_TX_EN;
	if( DATA0_1 & (1<<no) ){ d |= USB_TX_TOGL; }
	else{					d &= ~USB_TX_TOGL; }
	d |= USB_TX_LAST;
	write_reg(TXC[no], d);
	DATA0_1 ^= (1<<no);
}

/* M */
static void tx_enable(int no, bool last)
{
	byte cmd;
	cmd = USB_TX_EN;
	if( last ){ cmd |= USB_TX_LAST; }
	if( DATA0_1 & (1<<no) ){ cmd |= USB_TX_TOGL; }
	write_reg(TXC[no], cmd);
	DATA0_1 ^= (1<<no);
}

/*----- I/O ---------------------------------------------*/
/* dummy wait */
static void nop(void)
{
}
/* read */
static byte read_reg(byte adr)
{
	USB9604R = adr;
	nop();
	return USB9604D;
}
/* write */
static void write_reg(byte adr, byte data)
{
	USB9604R = adr;
	nop();
	USB9604D = data;
}

/* burst read */
static int read_reg_burst(byte adr, byte adr2, byte *buff,int cnt)
{
	int i;
	int rcnt;

	if( (rcnt = (read_reg(adr2) & 0xf)) == 0 ){
		return 0;
	}
	USB9604R = adr;
	for( i = 0; i < cnt; i++ ){
		*buff++ = USB9604D;
		if( --rcnt == 0 ){
			if( (rcnt = (read_reg(adr2) & 0xf)) == 0 ){
				i++;
				break;
			}
			USB9604R = adr;
		}
	}
	return i;
}

/* burst write */
static int write_reg_burst(byte adr, const byte *buff,int cnt)
{
	int i;
	int scnt;
	USB9604R = adr;
	for( scnt = 0,i = 0; i < cnt; i++ ){
		USB9604D = *buff;
		buff++;
		scnt++;
	}
	return scnt;
}
